package io.tswf.transmutations.staticview.generation.mapper

import groovy.transform.CompileStatic
import io.tswf.transmutations.staticview.exception.InvalidStaticViewMethodException
import io.tswf.transmutations.staticview.generation.IStaticViewMethodMapper
import io.tswf.transmutations.staticview.generation.StaticViewMethodMappingContext
import io.tswf.transmutations.staticview.util.TypeMatchingUtil
import org.objectweb.asm.commons.Method
import org.objectweb.asm.tree.InsnNode
import org.objectweb.asm.tree.MethodInsnNode
import org.objectweb.asm.tree.MethodNode
import org.objectweb.asm.tree.TypeInsnNode
import org.objectweb.asm.tree.VarInsnNode

import java.lang.reflect.Modifier

import static org.objectweb.asm.Opcodes.CHECKCAST
import static org.objectweb.asm.Opcodes.ILOAD
import static org.objectweb.asm.Opcodes.INVOKESTATIC
import static org.objectweb.asm.Opcodes.INVOKEVIRTUAL
import static org.objectweb.asm.Opcodes.IRETURN

@CompileStatic
abstract class AbstractMethodInvokeStaticViewMethodMapper implements IStaticViewMethodMapper {

    Method targetMethodDescription

    AbstractMethodInvokeStaticViewMethodMapper(Method targetMethodDescription) {
        this.targetMethodDescription = targetMethodDescription
    }

    @Override
    void handleStaticViewMethodMapping(StaticViewMethodMappingContext context) {
        def targetMethod = context.targetClass.getMethodByDescription(targetMethodDescription)
        def accessorMethod = context.generatedAccessMethodInTarget

        def targetMethodParameterTypes = targetMethod.parameterTypes
        def accessorMethodParameterTypes = accessorMethod.parameterTypes

        // Валидируем параметры метода-цели и метода-аксессора для того, чтобы убедиться, что вызов одного можно будет произвести из другого
        if (targetMethodParameterTypes.size() != accessorMethodParameterTypes.size()) {
            throw new InvalidStaticViewMethodException("target method ${targetMethod.javaFullName} located in ${context.targetClass.javaName} parameter types ${ targetMethodParameterTypes.map { it.className } } missmatching with accessor parameter types ${ accessorMethodParameterTypes.map { it.className } } in method ${accessorMethod.javaFullName}", context.viewMethodMeta)
        }

        targetMethodParameterTypes.eachNext(accessorMethodParameterTypes) { targetMethodParameterType, accessorMethodParameterType, parameterIndex ->
            if (TypeMatchingUtil.areTypesMismatching(accessorMethodParameterType, targetMethodParameterType)) {
                throw new InvalidStaticViewMethodException("target method ${targetMethod.javaFullName} ${parameterIndex+1}'st parameter type ${targetMethodParameterType.className} missmatching with ${accessorMethodParameterType.className} type of ${parameterIndex+1}'st parameter of accessor ${accessorMethod.javaFullName}", context.viewMethodMeta)
            }
        }

        applyAdditionalValidations(context, accessorMethod, targetMethod)

        appendMethodInvocation(context, accessorMethod, targetMethod)
    }

    void loadInvocationArguments(
            StaticViewMethodMappingContext context,
            MethodNode accessorMethod,
            MethodNode targetMethod
    ) {
        // Загружаем параметры метода аксессора на стек, кастуя их к целевому типу на случай, если передали Object. Для примитивов CHECKCAST не работает.
        targetMethod.parameterTypes.eachNext(accessorMethod.parameterTypes) { targetParameterType, accessorParameterType, parameterIndex ->
            accessorMethod.instructions << new VarInsnNode(accessorParameterType.getOpcode(ILOAD), parameterIndex)

            if (!targetParameterType.isPrimitive()) {
                accessorMethod.instructions << new TypeInsnNode(CHECKCAST, targetParameterType.internalName)
            }
        }
    }

    abstract void applyAdditionalValidations(
            StaticViewMethodMappingContext context,
            MethodNode accessorMethod,
            MethodNode targetMethod
    )

    abstract void appendMethodInvocation(
            StaticViewMethodMappingContext context,
            MethodNode accessorMethod,
            MethodNode targetMethod
    )
}
