package io.tswf.transmutations.staticview.generation.mapper

import groovy.transform.CompileStatic
import io.tswf.transmutations.staticview.exception.InvalidStaticViewMethodException
import io.tswf.transmutations.staticview.generation.StaticViewMethodMappingContext
import org.objectweb.asm.Type
import org.objectweb.asm.commons.Method
import org.objectweb.asm.tree.InsnNode
import org.objectweb.asm.tree.MethodInsnNode
import org.objectweb.asm.tree.MethodNode
import org.objectweb.asm.tree.TypeInsnNode

import static org.objectweb.asm.Opcodes.*

@CompileStatic
class ConstructorInvokeStaticViewMethodMapper extends AbstractMethodInvokeStaticViewMethodMapper {

    ConstructorInvokeStaticViewMethodMapper(Method targetMethodDescription) {
        super(targetMethodDescription)
    }

    @Override
    void applyAdditionalValidations(StaticViewMethodMappingContext context, MethodNode accessorMethod, MethodNode targetMethod) {
        // Сравниваем для тех же целей возвращаемые значения методов
        def targetMethodReturnType = targetMethod.returnType
        def accessorMethodReturnType = accessorMethod.returnType

        // Проверяем, что тип статического аксессора равен типу создаваемого класса
        if (accessorMethodReturnType.internalName != context.targetClass.name) {
            throw new InvalidStaticViewMethodException("accesor method ${accessorMethod.javaFullName} must have return type ${context.targetClass.javaName}, but was ${accessorMethodReturnType.className}", context.viewMethodMeta)
        }

        // Проверяем, что целевой метод - конструктор
        if (targetMethod.name != "<init>") {
            throw new InvalidStaticViewMethodException("target method name must be '<init>', not a ${targetMethod.name} in ${targetMethod.javaFullName}!", context.viewMethodMeta)
        }

        // Проверяем, что у этого конструктора каким-то образом не нашлось не-void дескриптора
        if (targetMethodReturnType != Type.VOID_TYPE) {
            throw new InvalidStaticViewMethodException("target method ${targetMethod.javaFullName} return type must be a void, not a ${targetMethodReturnType.className}", context.viewMethodMeta)
        }
    }

    @Override
    void appendMethodInvocation(StaticViewMethodMappingContext context, MethodNode accessorMethod, MethodNode targetMethod) {

        // Создаем инстанс на стеке и дублируем его дважды
        accessorMethod.instructions << new TypeInsnNode(NEW, context.targetClass.name)
        accessorMethod.instructions << new InsnNode(DUP)
        accessorMethod.instructions << new InsnNode(DUP)

        // Загружаем параметры для вызова конструктора
        loadInvocationArguments(context, accessorMethod, targetMethod)

        // Вызываем конструктор
        accessorMethod.instructions << new MethodInsnNode(INVOKESPECIAL, context.targetClass.name, targetMethodDescription.name, targetMethodDescription.descriptor, false)

        // Отдаем полученный объект. Это всегда Object, так как мы только что создали его через конструктор. Таким образом можно сразу указать ARETURN, так короче.
        accessorMethod.instructions << new InsnNode(ARETURN)
    }
}
