package io.tswf.transmutations.staticview.generation.mapper

import groovy.transform.CompileStatic
import io.tswf.transmutations.staticview.exception.InvalidStaticViewMethodException
import io.tswf.transmutations.staticview.generation.IStaticViewMethodMapper
import io.tswf.transmutations.staticview.generation.StaticViewMethodMappingContext
import io.tswf.transmutations.staticview.util.TypeMatchingUtil
import org.objectweb.asm.Type
import org.objectweb.asm.tree.FieldInsnNode
import org.objectweb.asm.tree.InsnNode
import org.objectweb.asm.tree.TypeInsnNode
import org.objectweb.asm.tree.VarInsnNode

import java.lang.reflect.Modifier

import static org.objectweb.asm.Opcodes.CHECKCAST
import static org.objectweb.asm.Opcodes.GETSTATIC
import static org.objectweb.asm.Opcodes.ILOAD
import static org.objectweb.asm.Opcodes.IRETURN
import static org.objectweb.asm.Opcodes.PUTSTATIC

@CompileStatic
class FieldSetStaticViewMethodMapper implements IStaticViewMethodMapper {

    String fieldName
    Type fieldType

    FieldSetStaticViewMethodMapper(String fieldName, Type fieldType) {
        this.fieldName = fieldName
        this.fieldType = fieldType
    }

    @Override
    void handleStaticViewMethodMapping(StaticViewMethodMappingContext context) {
        def fieldToAccess = context.targetClass.getFieldByNameAndType(fieldName, fieldType)

        def accessor = context.generatedAccessMethodInTarget
        def accessorReturnType = accessor.returnType

        if (accessorReturnType != Type.VOID_TYPE) {
            throw new InvalidStaticViewMethodException("accessor method return type ${accessorReturnType.className} (in method ${accessor.javaFullName}) is not primitive void for setter. Setters can be only voids!", context.viewMethodMeta)
        }

        def isFieldToAccessStatic = Modifier.isStatic(fieldToAccess.access)

        if (Modifier.isStatic(fieldToAccess.access)) {
            fieldToAccess.access = Modifier.removeModifier(fieldToAccess.access, Modifier.FINAL)
        }

        if (!isFieldToAccessStatic) {
            throw new InvalidStaticViewMethodException("target field '${fieldName}: ${fieldType.className}' is not static", context.viewMethodMeta)
        }

        def accessorParameterTypes = accessor.parameterTypes

        if (accessorParameterTypes.size() != 1) {
            throw new InvalidStaticViewMethodException("the accessor method parameters count is invalid. Expected 1, but found ${accessorParameterTypes.size()} for ${accessor.javaFullName}", context.viewMethodMeta)
        }

        def accessorNewValueParameterType = accessorParameterTypes.first()

        if (TypeMatchingUtil.areTypesMismatching(accessorNewValueParameterType, fieldType)) {
            throw new InvalidStaticViewMethodException("value parameter of accessor is invalid. Expected ${fieldType.className}, but found ${accessorNewValueParameterType.className} in ${accessor.javaFullName}", context.viewMethodMeta)
        }

        accessor.instructions << new VarInsnNode(accessorNewValueParameterType.getOpcode(ILOAD), 0)
        accessor.instructions << new TypeInsnNode(CHECKCAST, fieldType.internalName)
        accessor.instructions << new FieldInsnNode(PUTSTATIC, context.targetClass.name, fieldName, fieldType.descriptor)
        accessor.instructions << new InsnNode(accessorReturnType.getOpcode(IRETURN))
    }
}
