package io.tswf.transmutations.staticview.reading.mapper

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import io.tswf.transmutations.common.remapping.TransmutationRemapper
import io.tswf.transmutations.staticview.annotation.ConstructorAccessor
import io.tswf.transmutations.staticview.exception.InvalidStaticViewMethodException
import io.tswf.transmutations.staticview.generation.mapper.ConstructorInvokeStaticViewMethodMapper
import io.tswf.transmutations.staticview.reading.IStaticViewMethodMapperReader
import io.tswf.transmutations.staticview.reading.StaticViewMethodMapperReadingContext
import io.tswf.transmutations.common.annotation.ClassReference
import io.tswf.transmutations.common.reader.ClassReferenceReader
import org.objectweb.asm.Type
import org.objectweb.asm.commons.Method
import org.objectweb.asm.tree.AnnotationNode

@CompileStatic
@AutoService(IStaticViewMethodMapperReader)
class ConstructorInvokeStaticViewMethodMapperReader implements IStaticViewMethodMapperReader {

    public static final String NAME = "newInstance"
    public static final String DESCRIPTOR_PARAMETER_NAME = "desc"
    public static final String PARAMETER_TYPES_PARAMETER_NAME = "parameterTypes"

    @Override
    void handleStaticViewMethodReading(StaticViewMethodMapperReadingContext context) {
        def isMethodMatchingByName = context.targetMethod.name == NAME
        def constructorAccessorAnnotation = context.targetMethod.findAnnotationByType(ConstructorAccessor)

        def isMethodConstructorAccessor = isMethodMatchingByName || constructorAccessorAnnotation != null

        if (isMethodConstructorAccessor) {
            def rawConstructorDescription

            if (constructorAccessorAnnotation != null) {
                def constructorParameterTypes = constructorAccessorAnnotation.findValueByName(PARAMETER_TYPES_PARAMETER_NAME, AnnotationNode[])
                    ?.map {
                        ClassReferenceReader.getTypeFromClassRefAnnotation(it)
                    }

                def constructorDescriptor = constructorAccessorAnnotation.findValueByName(DESCRIPTOR_PARAMETER_NAME, String)

                if (constructorDescriptor != null && constructorParameterTypes != null) {
                    throw new InvalidStaticViewMethodException("'${PARAMETER_TYPES_PARAMETER_NAME}' can not be used with '${DESCRIPTOR_PARAMETER_NAME}' parameter in @${constructorAccessorAnnotation.type.className}", context.viewMethodMeta)
                }

                rawConstructorDescription = new Method(
                        "<init>",
                        constructorDescriptor ?: createConstructorDescriptor(constructorParameterTypes)
                )
            }

            if (rawConstructorDescription == null) {
                rawConstructorDescription = new Method(
                        "<init>",
                        createConstructorDescriptor(context.targetMethod.parameterTypes)
                )
            }

            def mappedConstructorDescription = TransmutationRemapper.mapMethod(context.viewMethodMeta.view.targetType, rawConstructorDescription)

            context.result << new ConstructorInvokeStaticViewMethodMapper(mappedConstructorDescription)
        }
    }

    static String createConstructorDescriptor(Collection<Type> parameterTypes) {
        return "(${ parameterTypes.map { it.descriptor }.join("") })V"
    }
}

